import os
from contextvars import ContextVar
from typing import Optional

from typing_extensions import ContextManager

METRIC = "metric"
IMPERIAL = "imperial"


_unit_system: ContextVar[Optional[str]] = ContextVar("unit_system", default=None)

_default_units = {
    METRIC: {
        "Distance": "km",
        "Duration": "s",
        "Pace": "seconds_per_km",
        "Speed": "km_h",
    },
    IMPERIAL: {
        "Distance": "mi",
        "Duration": "s",
        "Pace": "seconds_per_mile",
        "Speed": "mph",
    },
}


def get_unit_system() -> str:
    unit_system = _unit_system.get()
    return unit_system or os.environ.get("PY42195_UNIT_SYSTEM", METRIC)  # type: ignore


def set_unit_system(system: str) -> ContextManager:
    if system not in [METRIC, IMPERIAL]:
        raise ValueError(f"Invalid unit system: {system}")

    class _UnitSystem:
        def __enter__(self):
            pass

        def __exit__(self, exc_type, exc_val, exc_tb):
            _unit_system.reset(_reset_token)

    _reset_token = _unit_system.set(system)
    return _UnitSystem()


def get_default_unit(quantity: type) -> str:
    """Name of the argument that will be used as default unit for the given quantity."""
    return _default_units[get_unit_system()][quantity.__name__]



import pytest

from py42195.config import IMPERIAL, METRIC, get_unit_system, set_unit_system


@pytest.fixture(autouse=True)
def clean_env(monkeypatch):
    monkeypatch.delenv("PY42195_UNIT_SYSTEM", raising=False)
    yield


def test_default_is_metric():
    assert get_unit_system() == METRIC


def test_os_env_overrides_default(monkeypatch):
    monkeypatch.setenv("PY42195_UNIT_SYSTEM", IMPERIAL)
    assert get_unit_system() == IMPERIAL


def test_context_manager():
    assert get_unit_system() == METRIC
    with set_unit_system(IMPERIAL):
        assert get_unit_system() == IMPERIAL
    assert get_unit_system() == METRIC



from py42195.config import IMPERIAL, METRIC, get_unit_system, set_unit_system
from py42195.types import (
    Distance,
    Duration,
    Pace,
    distance,
    duration,
    pace,
    Speed,
    speed,
)

__all__ = [
    "Distance",
    "Duration",
    "Pace",
    "Speed",
    "distance",
    "duration",
    "pace",
    "speed",
    "set_unit_system",
    "get_unit_system",
    "METRIC",
    "IMPERIAL",
]



from py42195.utils import parse_interval
from py42195.types import Duration

import pytest
from datetime import timedelta


class TestParseInterval:
    @pytest.mark.parametrize(
        ("source", "expected"),
        [
            ("1", 1),
            ("1:12", 72),
            ("1:12:12", 4332),
            ("0.1", 0.1),
            ("999:59:59", 3599999),
        ],
    )
    def test_valid_string(self, source, expected):
        assert parse_interval(source) == timedelta(seconds=expected)

    @pytest.mark.parametrize(
        "source",
        [
            "2:",
            "1:74",
            "1:1:74",
            "1.2.5",
            "12km",
            "abc 1:12:12",
            "1:20pm",
        ],
    )
    def test_invalid_string(self, source):
        with pytest.raises(ValueError):
            parse_interval(source)

    @pytest.mark.parametrize("source", [Duration(1), 12])
    def test_invalid_type(self, source):
        with pytest.raises(TypeError):
            parse_interval(source)



import math
import re
from datetime import timedelta

INTERVAL_PATTERN = r"((?P<hour>\d+:)?(?P<minute>\d?\d:))?(?P<second>\d?\d(\.\d+)?)"


def parse_interval(source: str, /) -> timedelta:
    """Parse various intervals that can represent duration.

    :param s: Interval in the "[hh]:[mm]:ss[.sss]" format
    """
    if not isinstance(source, str):
        raise TypeError(f"Expected string, got {type(source)}")
    if not (match := re.match(f"^{INTERVAL_PATTERN}$", source)):
        raise ValueError(f"Cannot parse as time: {source}")

    h, m = (int(match[i][:-1]) if match[i] else 0 for i in ("hour", "minute"))
    s = float(match["second"])

    if (h or m) and s >= 60:
        raise ValueError(f"Cannot parse as time: {source}")
    if h and m >= 60:
        raise ValueError(f"Cannot parse as time: {source}")

    return timedelta(hours=h, minutes=m, seconds=s)


def format_interval(
    interval: timedelta | float,
    /,
    *,
    # TODO: Change into format?
    int_seconds: bool = False,
) -> str:
    if isinstance(interval, timedelta):
        # TODO: Some other delta type?
        interval = interval.total_seconds()
    if not math.isfinite(interval):
        return "-"
    m, s = divmod(interval, 60)
    sec_text = str(int(s)) if int_seconds else f"{s:.1f}"
    if m >= 60:
        h, m = divmod(m, 60)
        return f"{int(h)}:{'0' if m < 10 else ''}{int(m)}:{'0' if s < 10 else ''}{sec_text}"
    else:
        return f"{int(m)}:{'0' if s < 10 else ''}{sec_text}"



import math
from datetime import timedelta

import pytest

from py42195.config import IMPERIAL, METRIC, set_unit_system
from py42195.types import Distance, Pace, Speed, distance, duration, pace, speed


@pytest.fixture
def use_imperial_units():
    with set_unit_system(IMPERIAL):
        yield


class TestDistance:
    def test_marathon_distances(self):
        assert pytest.approx(Distance.MARATHON.km, abs=0.01) == 42.195
        assert pytest.approx(Distance.HALF_MARATHON.mi, abs=0.01) == 13.11

    @pytest.mark.parametrize(
        ("source", "expected"),
        [
            ("1.4", 1.4),
            ("1 km", 1),
            ("2mi", 3.218),
            ("750 m", 0.75),
            ("-4km", ValueError),
            ("1760yd", 1.609),
            ("10ft", 0.003048),
            ("m48", ValueError),
            ("a lot", ValueError),
        ],
    )
    def test_parse(self, source, expected):
        if isinstance(expected, type) and issubclass(expected, Exception):
            with pytest.raises(expected):
                Distance.parse(source)
        else:
            assert Distance.parse(source).km == pytest.approx(expected, abs=0.001)

    @pytest.mark.parametrize(
        ("kwargs", "error"),
        [
            ({"m": 1, "km": 2}, ValueError),
            ({"aa": 1}, TypeError),
            ({}, ValueError),
            ({"km": "many"}, TypeError),
        ],
    )
    def test_invalid_args(self, kwargs, error):
        with pytest.raises(error):
            Distance(**kwargs)

    @pytest.mark.parametrize(
        ("unit_system", "value", "expected"),
        [
            (METRIC, "1 km", "1.00 km"),
            (IMPERIAL, "1 km", "0.62 mi"),
            (None, "1 mi", "1.61 km"),
        ],
    )
    def test_str(self, unit_system, value, expected):
        a_distance = distance(value)
        with set_unit_system(unit_system or METRIC):
            assert str(a_distance) == expected

    def test_rtruediv(self):
        duration = timedelta(seconds=240)
        distance = Distance(km=1)
        assert duration / distance == pace("4:00")


class TestPace:
    class TestParse:
        @pytest.mark.parametrize(
            ("source", "expected"),
            [
                ("4:00", 240),
                ("4:00/km", 240),
                ("4:00 /mi", 149.1),
                ("4:00.0", 240),
                ("3:14.5", 194.5),
                ("1:00:01", 3601),
            ],
        )
        def test_valid(self, source, expected):
            pace = Pace.parse(source)
            assert pace.seconds_per_km == pytest.approx(expected, abs=0.1)

        @pytest.mark.parametrize(
            "source",
            [
                "1:74",
                "1 km",
                "fast",
                "1 s",
                "1:45:45:10",
                "1.1:2",
                "45.2:12",
                "1:69:02",
                "1::",
            ],
        )
        def test_invalid(self, source):
            with pytest.raises(ValueError):
                Pace.parse(source)

        def test_parse_with_imperial(
            self,
            use_imperial_units,
        ):
            pace = Pace.parse("4:00")
            assert pace.seconds_per_mile == pytest.approx(240)

    @pytest.mark.parametrize(
        ("seconds_per_km", "seconds_per_mile"), [(240, 386.24), (math.inf, math.inf)]
    )
    def test_units_equivalence(self, seconds_per_km, seconds_per_mile):
        pace_km = Pace(seconds_per_km=seconds_per_km)
        pace_mi = Pace(seconds_per_mile=seconds_per_mile)
        assert pace_km.seconds_per_km == pytest.approx(pace_mi.seconds_per_km, abs=0.01)

    class TestToSpeed:
        @pytest.mark.parametrize(
            ("a_speed", "a_pace"),
            [
                (10, "6:00"),
                ("10 mph", "3:43.7"),
                ("5 m/s", "3:20"),
            ],
        )
        def test_equivalence(self, a_speed, a_pace):
            assert speed(a_speed).km_h == pytest.approx(
                Pace.parse(a_pace).to_speed().km_h, abs=0.001
            )

    def test_rtruediv(self):
        duration = timedelta(seconds=240)
        pace = Pace(seconds_per_km=480)
        assert duration / pace == Distance(km=0.5)


class TestSpeed:
    class TestParse:
        @pytest.mark.parametrize(
            ("source", "expected"),
            [("10", 10), ("8 km/h", 8), ("1 mi/h", 1.609), ("2 m/s", 7.2)],
        )
        def test_valid(self, source, expected):
            speed = Speed.parse(source)
            assert speed.km_h == pytest.approx(expected, abs=0.01)

    class TestToPace:
        @pytest.mark.parametrize(
            ("a_speed", "a_pace"),
            [
                (10, "6:00"),
                ("10 mph", "3:43.7"),
                ("5 m/s", "3:20"),
            ],
        )
        def test_equivalence(self, a_speed, a_pace):
            assert speed(a_speed).to_pace().seconds_per_km == pytest.approx(
                Pace.parse(a_pace).seconds_per_km, abs=0.1
            )


class TestArithmetics:
    def test_marathon_record_pace(self):
        pace = duration("2:00:35") / Distance.MARATHON
        assert str(pace) == "2:51.5/km"

    def test_marathon_at_4_min_pace(self):
        duration = Distance.MARATHON * pace("4:00")
        assert str(duration) == "2:48:46.8"

    def test_sum(self):
        assert sum([Distance(km=1), Distance(km=3)]) == Distance(km=4)



import re
from datetime import timedelta
from functools import total_ordering
from typing import Any, Self

from typing_extensions import ClassVar, Optional

from py42195.config import get_default_unit, get_unit_system
from py42195.constants import (
    FEET_IN_KM,
    HALF_MARATHON_IN_KM,
    M_IN_KM,
    MARATHON_IN_KM,
    MILES_IN_KM,
    YARDS_IN_KM,
)
from py42195.utils import INTERVAL_PATTERN, format_interval, parse_interval


@total_ordering
class Distance:
    km: float

    MARATHON: ClassVar["Distance"]
    HALF_MARATHON: ClassVar["Distance"]

    ALLOWED_UNITS = ["km", "mi", "m", "yd", "ft"]
    PARSE_PATTERN: re.Pattern = re.compile(
        r"^(?P<value>\d+(\.\d+)?)\s*(?P<unit>" + "|".join(ALLOWED_UNITS) + ")?$"
    )

    def __init__(
        self,
        *,
        km: Optional[float] = None,
        m: Optional[float] = None,
        mi: Optional[float] = None,
        yd: Optional[float] = None,
        ft: Optional[float] = None,
    ):
        args_given = [
            unit
            for unit, val in locals().items()
            if unit in self.ALLOWED_UNITS and val is not None
        ]
        if len(args_given) != 1:
            raise ValueError(
                f"Exactly one of the following arguments should be provided: {', '.join(self.ALLOWED_UNITS)}, "
                f"got {len(args_given)}: {', '.join(args_given)}"
            )

        if km is not None:
            self.km = km
        elif m is not None:
            self.km = m * M_IN_KM
        elif yd is not None:
            self.km = yd * YARDS_IN_KM
        elif mi is not None:
            self.km = mi * MILES_IN_KM
        elif ft is not None:
            self.km = ft * FEET_IN_KM

        if not isinstance(self.km, (int, float)):
            raise TypeError(f"Expected a number, got {type(self.km)}")

    @property
    def mi(self) -> float:
        return self.km / MILES_IN_KM

    @classmethod
    def parse(cls, s: str) -> Self:
        # split into unit and value
        match = cls.PARSE_PATTERN.match(s)
        if match is None:
            raise ValueError(f"Cannot parse as distance: {s}")
        value = float(match.group("value"))
        unit = match.group("unit") or "km"
        return cls(**{unit: value})

    def __str__(self) -> str:
        if get_unit_system() == "imperial":
            return f"{self.mi:.2f} mi"
        return f"{self.km:.2f} km"

    def __repr__(self) -> str:
        if get_unit_system() == "imperial":
            return f"Distance(mi={self.mi})"
        return f"Distance(km={self.km})"

    def __add__(self, other: "Distance") -> "Distance":
        if isinstance(other, Distance):
            return Distance(km=self.km + other.km)
        return NotImplemented

    def __radd__(self, other: Any) -> "Distance":
        if other == 0:
            # Support sum
            return self
        return other + self

    def __sub__(self, other: "Distance") -> "Distance":
        if isinstance(other, Distance):
            return Distance(km=self.km - other.km)
        return NotImplemented

    def __mul__(self, other):
        if isinstance(other, (int, float)):
            return Distance(km=self.km * other)
        if isinstance(other, Pace):
            return Duration(self.km * other.seconds_per_km)
        else:
            return NotImplemented

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if isinstance(other, (int, float)):
            return Distance(km=self.km / other)
        if isinstance(other, Distance):
            return self.km / other.km
        if isinstance(other, Duration):
            return Speed(km_h=self.km / (other.seconds * 3600))
        if isinstance(other, Speed):
            return Duration(self.km / other.km_h * 3600)
        return NotImplemented

    def __rtruediv__(self, other):
        if isinstance(other, timedelta):
            return Duration(other) / self
        return NotImplemented

    def __eq__(self, other: object, /) -> bool:
        if isinstance(other, Distance):
            return self.km == other.km
        return NotImplemented

    def __lt__(self, other: object, /) -> bool:
        if isinstance(other, Distance):
            return self.km < other.km
        return NotImplemented


Distance.MARATHON = Distance(km=MARATHON_IN_KM)
Distance.HALF_MARATHON = Distance(km=HALF_MARATHON_IN_KM)


@total_ordering
class Duration:
    seconds: float

    def __init__(self, seconds: float | timedelta, /):
        if isinstance(seconds, timedelta):
            self.seconds = seconds.total_seconds()
        else:
            self.seconds = seconds

    @classmethod
    def parse(cls, s: str, /) -> Self:
        if (delta := parse_interval(s)) is None:
            raise ValueError(f"Cannot parse as time: {s}")
        return cls(delta.total_seconds())

    def __str__(self) -> str:
        return format_interval(self.seconds)

    def __repr__(self) -> str:
        return f"duration('{self!s}')"

    def __add__(self, other: "Duration") -> "Duration":
        if isinstance(other, Duration):
            return Duration(self.seconds + other.seconds)
        return NotImplemented

    def __radd__(self, other: Any) -> "Duration":
        if other == 0:
            # Support sum
            return self
        return other + self

    def __sub__(self, other: "Duration") -> "Duration":
        if isinstance(other, Duration):
            return Duration(self.seconds - other.seconds)
        return NotImplemented

    def __mul__(self, other: Any) -> Any:
        if isinstance(other, (int, float)):
            return Duration(self.seconds * other)
        if isinstance(other, Speed):
            return Distance(km=self.seconds * other.km_h / 3600)
        return NotImplemented

    def __truediv__(self, other):
        if isinstance(other, (int, float)):
            return Duration(self.seconds / other)
        if isinstance(other, Distance):
            return Pace(seconds_per_km=self.seconds / other.km)
        if isinstance(other, Pace):
            return Distance(km=self.seconds / other.seconds_per_km)
        if isinstance(other, Duration):
            return self.seconds / other.seconds
        if isinstance(other, timedelta):
            return self.seconds / other.total_seconds()
        return NotImplemented

    def __eq__(self, other: object, /) -> bool:
        if isinstance(other, Duration):
            return self.seconds == other.seconds
        if isinstance(other, timedelta):
            return self.seconds == other.total_seconds()
        return NotImplemented

    def __lt__(self, other: object, /) -> bool:
        if isinstance(other, Duration):
            return self.seconds < other.seconds
        if isinstance(other, timedelta):
            return self.seconds < other.total_seconds()
        return NotImplemented


@total_ordering
class Pace:
    seconds_per_km: float

    ALLOWED_UNITS = {
        "/km": "seconds_per_km",
        "/mi": "seconds_per_mile",
    }
    PARSE_PATTERN: re.Pattern = re.compile(
        f"^(?P<interval>{INTERVAL_PATTERN})\\s*(?P<unit>"
        + "|".join(ALLOWED_UNITS)
        + ")?$"
    )

    def __init__(
        self,
        *,
        seconds_per_km: Optional[float] = None,
        seconds_per_mile: Optional[float] = None,
    ):
        if (seconds_per_mile is not None) and (seconds_per_km is not None):
            raise ValueError(
                "Either seconds_per_mile or seconds_per_km should be provided"
            )
        elif seconds_per_km is not None:
            self.seconds_per_km = seconds_per_km
        elif seconds_per_mile is not None:
            self.seconds_per_km = seconds_per_mile / MILES_IN_KM
        else:
            raise ValueError(
                "Either seconds_per_mile or seconds_per_km should be provided"
            )

    @property
    def seconds_per_mile(self) -> float:
        return self.seconds_per_km * MILES_IN_KM

    def __str__(self) -> str:
        if get_unit_system() == "imperial":
            return format_interval(self.seconds_per_mile) + "/mi"
        return format_interval(self.seconds_per_km) + "/km"

    def __repr__(self) -> str:
        if get_unit_system() == "imperial":
            return f"Pace(seconds_per_mile={self.seconds_per_mile})"
        return f"Pace(seconds_per_km={self.seconds_per_km})"

    def __add__(self, other):
        if isinstance(other, Pace):
            return Pace(seconds_per_km=self.seconds_per_km + other.seconds_per_km)
        return NotImplemented

    def __radd__(self, other: Any) -> "Pace":
        if other == 0:
            # Support sum
            return self
        return other + self

    def __sub__(self, other):
        if isinstance(other, Pace):
            return Pace(seconds_per_km=self.seconds_per_km - other.seconds_per_km)
        return NotImplemented

    def __mul__(self, other):
        if isinstance(other, (int, float)):
            return Pace(seconds_per_km=self.seconds_per_km * other)
        if isinstance(other, Distance):
            return Duration(self.seconds_per_km * other.km)
        else:
            return NotImplemented

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if isinstance(other, (int, float)):
            return Pace(seconds_per_km=self.seconds_per_km / other)
        if isinstance(other, Pace):
            return self.seconds_per_km / other.seconds_per_km
        return NotImplemented

    def __rtruediv__(self, other):
        if isinstance(other, timedelta):
            return Duration(other) / self
        return NotImplemented

    def __eq__(self, other: object, /) -> bool:
        if isinstance(other, Pace):
            return self.seconds_per_km == other.seconds_per_km
        return NotImplemented

    def __lt__(self, other: object, /) -> bool:
        if isinstance(other, Pace):
            return self.seconds_per_km < other.seconds_per_km
        return NotImplemented

    @classmethod
    def parse(cls, s: str, /) -> Self:
        match = re.match(cls.PARSE_PATTERN, s)
        if not match:
            raise ValueError(f"Cannot parse as pace: {s}")
        unit = cls.ALLOWED_UNITS.get(match.group("unit")) or get_default_unit(cls)
        value = parse_interval(match.group("interval")).total_seconds()
        return cls(**{unit: value})

    def to_speed(self) -> "Speed":
        return Speed(km_h=3600 / self.seconds_per_km)


@total_ordering
class Speed:
    km_h: float

    ALLOWED_UNITS = {
        "kmh": "km_h",
        "km/h": "km_h",
        "mph": "mph",
        "mi/h": "mph",
        "m/s": "m_s",
    }
    PARSE_PATTERN: re.Pattern = re.compile(
        r"^(?P<value>\d+(\.\d+)?)\s*(?P<unit>" + "|".join(ALLOWED_UNITS) + ")?$"
    )

    def __init__(
        self,
        *,
        km_h: Optional[float] = None,
        mph: Optional[float] = None,
        m_s: Optional[float] = None,
    ) -> None:
        args_given = [
            unit
            for unit, val in locals().items()
            if unit in self.ALLOWED_UNITS.values() and val is not None
        ]
        if len(args_given) != 1:
            raise ValueError(
                f"Exactly one of the following arguments should be provided: {', '.join(self.ALLOWED_UNITS)}, "
                f"got {len(args_given)}: {', '.join(args_given)}"
            )

        if km_h is not None:
            self.km_h = km_h
        elif mph is not None:
            self.km_h = mph * MILES_IN_KM
        elif m_s is not None:
            self.km_h = m_s * 3600 / 1000

        if not isinstance(self.km_h, (int, float)):
            raise ValueError("Speed should be a number.")

    def __lt__(self, other: object, /) -> bool:
        if isinstance(other, Speed):
            return self.km_h < other.km_h
        return NotImplemented

    def __eq__(self, other: object, /) -> bool:
        if isinstance(other, Speed):
            return self.km_h == other.km_h
        return NotImplemented

    @classmethod
    def parse(cls, s: str) -> Self:
        # split into unit and value
        match = cls.PARSE_PATTERN.match(s)
        if match is None:
            raise ValueError(f"Cannot parse as speed: {s}")
        value = float(match.group("value"))
        unit = cls.ALLOWED_UNITS.get(match.group("unit")) or "km_h"
        return cls(**{unit: value})

    @property
    def m_s(self) -> float:
        return self.km_h * 1000 / 3600

    @property
    def mph(self) -> float:
        return self.km_h / MILES_IN_KM

    def to_pace(self) -> Pace:
        return Pace(seconds_per_km=3600 / self.km_h)

    def __mul__(self, other):
        if isinstance(other, (int, float)):
            return Speed(km_h=self.km_h * other)
        if isinstance(other, Duration):
            return Distance(km=self.km_h * other.seconds / 3600)
        else:
            return NotImplemented

    def __rmul__(self, other):
        return self.__mul__(other)

    def __add__(self, other, /):
        if isinstance(other, Speed):
            return Speed(km_h=self.km_h + other.km_h)
        return NotImplemented

    def __radd__(self, other: Any) -> "Speed":
        if other == 0:
            # Support sum
            return self
        return other + self

    def __sub__(self, other, /):
        if isinstance(other, Speed):
            return Speed(km_h=self.km_h - other.km_h)
        return NotImplemented

    def __truediv__(self, other, /):
        if isinstance(other, (int, float)):
            return Speed(km_h=self.km_h / other)
        if isinstance(other, Speed):
            return self.km_h / other.km_h
        return NotImplemented

    def __str__(self):
        return f"{self.km_h:.2f} km/h"

    def __repr__(self) -> str:
        return f"speed('{self!s}')"


def pace(value: Any, **kwargs) -> Pace:
    if isinstance(value, str):
        if kwargs:
            raise ValueError("Cannot provide kwargs when parsing from string")
        return Pace.parse(value)
    else:
        default_unit = get_default_unit(Pace)
        return Pace(**(kwargs | {default_unit: value}))


def duration(value: Any) -> Duration:
    """Construct a Duration object from various types."""
    if isinstance(value, str):
        return Duration.parse(value)
    if isinstance(value, Duration):
        return value
    else:
        return Duration(value)


def distance(value: Any, **kwargs) -> Distance:
    if isinstance(value, str):
        return Distance.parse(value)
    if isinstance(value, Distance):
        return value
    else:
        default_unit = get_default_unit(Distance)
        return Distance(**(kwargs | {default_unit: value}))


def speed(value: Any, **kwargs) -> Speed:
    if isinstance(value, str):
        return Speed.parse(value)
    if isinstance(value, Speed):
        return value
    else:
        default_unit = get_default_unit(Speed)
        return Speed(**(kwargs | {default_unit: value}))
